#include <cnoid/SimpleController>
#include <cnoid/Joystick>
#include <fmt/format.h>

using namespace std;
using namespace cnoid;
using fmt::format;

class TurtleBot2JoystickController : public SimpleController
{
    static const int WHEEL_NUM = 2;
    const string wheelNames[WHEEL_NUM] = { "wheel_left", "wheel_right" };
    Link::ActuationMode actuationMode;
    Link* wheels[2];
    Joystick joystick;
    Body* body;
    
public:
    virtual bool initialize(SimpleControllerIO* io) override
    {
        ostream& os = io->os();
        // Body情報取得.
        body = io->body();

        // リンクのアクチュエーションモード設定
        actuationMode = Link::JOINT_TORQUE;
        // コントローラオプションの取得
        string option = io->optionString();

        if(!option.empty()){
            // コントローラオプションが空の場合
            if(option == "velocity" || option == "position"){
                // velocityかpositionが指定されている場合
                actuationMode = Link::JOINT_VELOCITY;
            } else if(option == "torque"){
                // torqueが指定されている場合
                actuationMode = Link::JOINT_TORQUE;
            } else {
                // 上記以外の場合
                os << format("Warning: Unknown option \"{}\".", option) << endl;
            }
        }

        for(int i = 0; i < WHEEL_NUM; ++i){
            // ホイールリンクを取得
            wheels[i] = body->link(wheelNames[i]);
            if(!wheels[i]){
                // リンクが取得できない場合
                os << format("{0} of {1} is not found.", wheelNames[i], body->name()) << endl;
                return false;
            }

            // ホイールのアクチュエーションモードを設定
            wheels[i]->setActuationMode(actuationMode);
            // ホイールに対する出力を有効化
            io->enableOutput(wheels[i]);
        }
        return true;
    }
    
    virtual bool control() override
    {
        // ジョイスティックの状態取得
        joystick.readCurrentState();
        double pos[2]; // ジョイスティックの変化量
        for(int i = 0; i < 2; ++i){
            // ジョイスティックの値取得
            pos[i] = joystick.getPosition(
                    i == 0 ? Joystick::L_STICK_H_AXIS : Joystick::L_STICK_V_AXIS);

            if(fabs(pos[i]) < 0.2){
                // 変化量が0.2未満の場合
                pos[i] = 0.0;
            }
        }

        if(actuationMode == Link::JOINT_VELOCITY){
            // アクチュエーションモードがvelocitの場合
            static const double K = 20.0;
            wheels[0]->dq_target() = K * (-pos[1] + pos[0]);
            wheels[1]->dq_target() = K * (-pos[1] - pos[0]);
        }

        return true;
    }
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TurtleBot2JoystickController)
