#include <cnoid/SimpleController>
#include <cnoid/Camera>
#include <fmt/format.h>

using namespace std;
using namespace cnoid;
using fmt::format;

class TurtleBot2TurnLineController : public SimpleController
{
    // TurtleBot2StopLineControllerと同様
    static const int WHEEL_NUM = 2;
    const string wheelNames[WHEEL_NUM] = { "wheel_left", "wheel_right" };
    Link::ActuationMode actuationMode;
    Link* wheels[2];
    Body* body;
    // カメラデバイス情報格納変数
    CameraPtr camera;
    // 前回の画像格納変数
    std::shared_ptr<const Image> prevImage;
    // 画像内のグレー、白、黄色の数の格納変数
    int cnt[3] = { 0, 0, 0 };
    const double d = 0.115;
    // タイムステップ格納変数
    double dt;
    // PID制御の定数
    const double Kp = 48.0;
    const double Ki = 0.0092;
    const double Kd = 0.00015;
    // 画像内の白線の目標値
    static const int TARGET = 10000;
    // 前回と現在の偏差値格納変数
    double diff_R[2] = { 0, 0 };
    double diff_L[2] = { 0, 0 };
    // 偏差の積分値格納変数
    double integral[2] = { 0, 0 };
    // 偏差の微分値格納変数
    double derivation[2] = { 0, 0 };

public:
    virtual bool initialize(SimpleControllerIO* io) override
    {
        // TurtleBot2JoystickControllerと同様
        ostream& os = io->os();
        // Body情報取得.
        body = io->body();

        // リンクのアクチュエーションモード設定
        actuationMode = Link::JOINT_TORQUE;
        // コントローラオプションの取得
        string option = io->optionString();

        if(!option.empty()){
            // コントローラオプションが空の場合
            if(option == "velocity" || option == "position"){
                // velocityかpositionが指定されている場合
                actuationMode = Link::JOINT_VELOCITY;
            } else if(option == "torque"){
                // torqueが指定されている場合
                actuationMode = Link::JOINT_TORQUE;
            } else {
                // 上記以外の場合
                os << format("Warning: Unknown option \"{}\".", option) << endl;
            }
        }

        for(int i = 0; i < WHEEL_NUM; ++i){
            // ホイールリンクを取得
            wheels[i] = body->link(wheelNames[i]);
            if(!wheels[i]){
                // リンクが取得できない場合
                os << format("{0} of {1} is not found.", wheelNames[i], body->name()) << endl;
                return false;
            }

            // ホイールのアクチュエーションモードを設定
            wheels[i]->setActuationMode(actuationMode);
            // ホイールに対する出力を有効化
            io->enableOutput(wheels[i]);
        }
        
        // LineTraceカメラを取得
        camera = body->findDevice<Camera>("LineTrace");
        // カメラのコントローラへの入力を有効化
        io->enableInput(camera);
        // タイムステップの設定
        dt = io->timeStep();
        
        return true;
    }
    
    virtual bool control() override
    {
        double vx, va;
        va = 0.5;
        vx = 0.3;

        onCameraStateChanged();

        // 前回の偏差値を設定
        diff_L[0] = diff_L[1];
        // 現在の偏差値(目標値 - センサ値)を取得
        diff_L[1] = TARGET - cnt[1];
        // 偏差の積分値を取得。偏差の積分値 = (( 最新の偏差 + 前回の偏差 ) / 2 ) * 時間
        integral[0] += (diff_L[1] + diff_L[0]) / 2.0 * dt;
        // 偏差の微分値を取得。偏差の微分値 = ( 最新の偏差 - 前回の偏差 ) / 時間
        derivation[0] = (diff_L[1] - diff_L[0]) / dt;
        diff_R[0] = diff_R[1];
        // 現在の偏差値(目標値 - センサ値)を取得
        diff_R[1] = TARGET - cnt[1];
        integral[1] += (diff_R[1] + diff_R[0]) / 2.0 * dt;
        derivation[1] = (diff_R[1] - diff_R[0]) / dt;

        if(actuationMode == Link::JOINT_VELOCITY){
            // 関節速度の指令値格納変数
            double dq_target[2];

            // PID制御
            dq_target[0] = Kp * (vx - (va * d * diff_L[1] / 2500)) + Ki * (integral[0] * va * d) + Kd * (derivation[0] * va * d);
            dq_target[1] = Kp * (vx + (va * d * diff_R[1] / 2500)) + Ki * (integral[1] * va * d) + Kd * (derivation[1] * va * d);

            // 左右のホイールに指令値を与える
            wheels[0]->dq_target() = dq_target[0];
            wheels[1]->dq_target() = dq_target[1];
        }
        return true;
    }
    
    virtual void stop() override
    {
        for(int i = 0; i < WHEEL_NUM; ++i){
            diff_R[i] = diff_L[i] = 0;
            integral[i] = derivation[i] = 0;
        }

        for(int i = 0; i < 3; ++i){
            cnt[i] = 0;
        }
    }

    void onCameraStateChanged()
    {
        size_t length = 0;
        if(camera->sharedImage() != prevImage){
            // カメラ画像が更新されたか確認
            const Image& image = camera->constImage();
            if(!image.empty()){
                // カメラ画像が取得できた場合
                int width, height;
                // 画像のサイズを取得
                height = image.height();
                width = image.width();
                length = width * height * image.numComponents() * sizeof(unsigned char);
            }

            // 画像の1ピクセルごとのデータを取得
            unsigned char* src = (unsigned char*)image.pixels();

            // グレー、白、黄色のカウント用配列の初期化
            cnt[0] = cnt[1] = cnt[2] = 0;
            // RGB値格納配列
            int rgb[3];

            // データ数分ループ
            for(int i = 0; i < length / 3; ++i){
                // RGBの値を格納
                rgb[0] = (int)src[i * 3];
                rgb[1] = (int)src[i * 3 + 1];
                rgb[2] = (int)src[i * 3 + 2];

                if((rgb[0] >= 100 and rgb[0] < 180)
                        and (rgb[1] >= 100 and rgb[1] < 180)
                        and (rgb[2] >= 100 and rgb[2] < 180)
                        and abs(rgb[0] - rgb[1]) <= 10
                        and abs(rgb[1] - rgb[2]) <= 10
                        and abs(rgb[2] - rgb[0]) <= 10){
                    cnt[0]++;        // グレーの個数をカウント
                }else if(rgb[0] >= 180 and rgb[1] >= 180 and rgb[2] >= 180){
                    cnt[1]++;        // 白の個数をカウント
                }else if(rgb[0] >= 170 and rgb[1] >= 170 and rgb[2] <= 100){
                    cnt[2]++;        // 黄色の個数をカウント
                }
            }
            // 前回値の更新
            prevImage = camera->sharedImage();
        }
    }
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TurtleBot2TurnLineController)
