#include <cnoid/SimpleController>
#include <cnoid/RootItem>
#include <cnoid/SimulatorItem>
#include <fmt/format.h>

using namespace std;
using namespace cnoid;
using fmt::format;

class TurtleBot2TurningController : public SimpleController
{
    static const int WHEEL_NUM = 2;
    const string wheelNames[WHEEL_NUM] = { "wheel_left", "wheel_right" };
    Link::ActuationMode actuationMode;
    Link* wheels[2];
    Body* body;
    SimulatorItem* simItem;
    double startTime = 0.0;
    const double d = 0.115;
    const double Kp = 48.0;

public:
    virtual bool initialize(SimpleControllerIO* io) override
    {
        ostream& os = io->os();
        // Body情報取得.
        body = io->body();

        // リンクのアクチュエーションモード設定
        actuationMode = Link::JOINT_TORQUE;
        // コントローラオプションの取得
        string option = io->optionString();

        if(!option.empty()){
            // コントローラオプションが空の場合
            if(option == "velocity" || option == "position"){
                // velocityかpositionが指定されている場合
                actuationMode = Link::JOINT_VELOCITY;
            } else if(option == "torque"){
                // torqueが指定されている場合
                actuationMode = Link::JOINT_TORQUE;
            } else {
                // 上記以外の場合
                os << format("Warning: Unknown option \"{}\".", option) << endl;
            }
        }

        for(int i = 0; i < WHEEL_NUM; ++i){
            // ホイールリンクを取得
            wheels[i] = body->link(wheelNames[i]);
            if(!wheels[i]){
                // リンクが取得できない場合
                os << format("{0} of {1} is not found.", wheelNames[i], body->name()) << endl;
                return false;
            }

            // ホイールのアクチュエーションモードを設定
            wheels[i]->setActuationMode(actuationMode);
            // ホイールに対する出力を有効化
            io->enableOutput(wheels[i]);
        }
        
        // 開始時間の初期化
        startTime = 0.0;
        
        return true;
    }
    
    virtual bool start() override
    {
        // TurtleBot2という名称のアイテムを取得
        Item* robotItem = RootItem::instance()->findItem("TurtleBot2");
        // 実行中のシミュレータアイテムを取得
        simItem = SimulatorItem::findActiveSimulatorItemFor(robotItem);

        return true;
    }

    virtual bool control() override
    {
        // 車体の中心の速度vx(m/s), 旋回角速度va(rad/s)
        double vx, va;
        va = 0.0;
        vx = 0.3;

        if(simItem->simulationTime() < 2.0){
            // シミュレーション時間が2s未満の場合、直進
            va = 0.0;
            vx = 0.3;
        }else if(simItem->simulationTime() < 2.2){
            va = vx = 0.0;
        }else if(simItem->simulationTime() < 2.7){
            // 90°右旋回
            va = -2.41;
            vx = 0.0;
        }else if(simItem->simulationTime() < 2.9){
            va = vx = 0.0;
        }else if(simItem->simulationTime() < 4.9){
            // 2s間直進
            va = 0.0;
            vx = 0.3;
        }else if(simItem->simulationTime() < 5.1){
            va = vx = 0.0;
        }else if(simItem->simulationTime() < 5.6){
            // 90°左旋回
            va = 2.55;
            vx = 0.0;
        }else if(simItem->simulationTime() < 5.8){
            va = vx = 0.0;
        }else if(simItem->simulationTime() < 7.8){
            // 2s間直進
            va = 0.0;
            vx = 0.3;
        }else{
            // 停止
            va = vx = 0.0;
        }
        
        if(actuationMode == Link::JOINT_VELOCITY){
            // アクチュエーションモードがvelocityの場合
            // 関節速度の指令値格納変数
            double dq_target[2];

            if(startTime == 0.0){
                // 開始時間が0.0の場合
                // 開始時間に現在のシミュレーション時間を設定
                startTime = simItem->simulationTime();

            }
            
            dq_target[0] = Kp * (vx - va * d);
            dq_target[1] = Kp * (vx + va * d);
            // 左右のホイールに指令値を与える
            wheels[0]->dq_target() = dq_target[0];
            wheels[1]->dq_target() = dq_target[1];
        }

        return true;
    }
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TurtleBot2TurningController)

